//-------------------------------------------------------------------------
//An extensible, easy to use OpenGL wrapper
//Written by Ananth B. AKA HellSpawn
//-------------------------------------------------------------------------
unit GLClass;

interface

uses
    Forms,
    Classes,
    Windows,
    SysUtils,
    GL,
    GLU;

const
     //Whether we want double buffering or not
     DoubleBuffer:boolean=true;
     //Do we want color index mode?
     ColorIndex:boolean=false;
     //What is the color depth we need?
     ColorBits:byte=16;
     //How many bits per pixel should the ZBuffer have?
     DepthBits:byte=16;
     //How many bits per pixel should the stencil buffer have?
     StencilBits:byte=8;

type
    //This will denote a floating point color quad
    PColorf=^TColorf;
    TColorf=record
                  R,G,B,A:single;
    end;
    //Our class declaration
    TOpenGL=class(TObject)
            private
            //The device context
            fDC:HDC;
            //The OpenGL rendering context that we will init
            fHRC:HGLRC;
            //Palette in case the user needs color-index mode
            fPalette:HPALETTE;

            //The color using which the screen will be cleared
            fClearColor:TColorf;

            //Fullscreen? YES! YES!
            fFullScreen:boolean;
            //Resolution at which to go fullscreen
            fXRes,fYRes:integer;
            //If OpenGL inited using a form, a pointer to it
            fOwner:TForm;
            //Was OpenGL inited from just a handle?
            fFromHandle:boolean;
            //The handle from which OpenGL was inited
            fHandle:THandle;
            protected
            //Make our rendering context the current one.
            procedure MakeCurrent;

            //Access methods for some properties
            procedure SetClearColor(aColor:TColorf);
            procedure SetFullScreen(v:boolean);

            public
            //Sets up basic info.
            constructor Create(aOwner:TForm);
            //DUH!
            destructor Destroy;override;

            //Initialize OpenGL with the given info
            procedure InitGL;
            //Initialize OpenGL with the given info
            procedure InitGLFromHandle(aHandle:THandle);
            //Clean up after OpenGL
            procedure CleanUpGL;

            //Flip the backbuffer to the front buffer
            procedure Flip;
            //Clear all the inited buffer
            procedure ClearAll;
            //Clear the ZBuffer alone
            procedure ClearZBuffer;
            //Clear the Color buffer alone
            procedure ClearColorBuffer;
            //Clear the stencil buffer alone
            procedure ClearStencilBuffer;
            //The device context
            property DC:HDC read fDC;
            //The rendering context
            property RC:HGLRC read fHRC;
            //The palette
            property Palette:HPALETTE
                     read fPalette
                     write fPalette;
            //The clear color
            property ClearColor:TColorf
                     read fClearColor
                     write SetClearColor;
            //An alien from mars...
            //Just checking to see if you are really reading these silly
            //comments
            property XRes:integer
                     read fXRes
                     write fXRes
                     default 640;
            //Y resolution
            property YRes:integer
                     read fYRes
                     write fYRes
                     default 480;
            //FullScreen property
            property FullScreen:boolean
                     read fFullScreen
                     write SetFullScreen;
    end;
    //A **really** simple class, but goddamn helpful around the place
    TViewVolume=class
                     private
                     //The field-of-view
                     fFOV,
                     //distance to the near clipping plane
                     fHither,
                     //distance to the far clipping plane
                     fYon:double;
                     //Viewport information
                     fX,fY,
                     fWidth,
                     fHeight:integer;
                     public
                     //I really should stop commenting each line
                     constructor Create;
                     //Set the projection params
                     procedure SetMatrix;
                     //Properties for the private member vars
                     //described above
                     property FOV:double
                              read fFOV
                              write fFOV;
                     property Hither:double
                              read fHither
                              write fHither;
                     property Yon:double
                              read fYon
                              write fYon;
                     property X:integer
                              read fX
                              write fX;
                     property Y:integer
                              read fY
                              write fY;
                     property Width:integer
                              read fWidth
                              write fWidth;
                     property Height:integer
                              read fHeight
                              write fHeight;

    end;
//Very helpful functions.
//Both as far as the wrapper and general OpenGL stuff

//Change the desktop resolution and colordepth arbitrarily
procedure ChangeRes(rw,rh,rbpp:cardinal);
//Get the vendor string of the video card
function GetVendor:string;
//Get the current machine's renderer
function GetRenderer:string;
//Get the OpenGL version
function GetGLVersion:string;
//Get the GLU version
function GetGLUVersion:string;
//Here comes the interesting part

//Fetch the list of all extensions that the video card supports
function GetExtensions:string;
//Get an extension by number. Extensions start at one
//according to my tokenizer
function GetExtension(num:integer):string;
//Get the Number of tokens.
function GetNumExtensions:integer;

implementation

//This is a simple procedure that, given the device context
//that we want to render to, initializes the DC (and palette)
procedure InitDC(DC: HDC; Palette: HPalette);
var
  hHeap: Integer;
  nColors, i: Integer;
  lpPalette: PLogPalette;
  byRedMask, byGreenMask, byBlueMask: Byte;
  nPixelFormat: Integer;
  pfd: TPixelFormatDescriptor;
begin
      // Initialise the form's DC for OpenGL, and setup a palette.
      FillChar(pfd, SizeOf(pfd), 0);
      with pfd do
      begin
           nSize:=SizeOf(TPixelFormatDescriptor);// Size of this structure
           nVersion:= 1;// Version number
           dwFlags:=PFD_DRAW_TO_WINDOW or
                    PFD_SUPPORT_OPENGL;// Flags
           if DoubleBuffer then
             dwFlags:=dwFlags or PFD_DOUBLEBUFFER;// Double buffering?
           if ColorIndex then iPixelType:=PFD_TYPE_COLORINDEX// Color index pixel values
           else iPixelType:=PFD_TYPE_RGBA;// RGBA pixel values
           cColorBits:=ColorBits;
           cDepthBits:=DepthBits;
           cStencilBits:=StencilBits;
           iLayerType:=PFD_MAIN_PLANE;// Layer type
      end;
      nPixelFormat:=ChoosePixelFormat(DC,@pfd);
      SetPixelFormat(DC, nPixelFormat,@pfd);

      DescribePixelFormat(DC,nPixelFormat,SizeOf(TPixelFormatDescriptor),pfd);

      if ((pfd.dwFlags and PFD_NEED_PALETTE)<>0) then
      begin
           nColors:=1 shl pfd.cColorBits;
           hHeap:=GetProcessHeap;
           lpPalette:=HeapAlloc(hHeap, 0, sizeof(TLogPalette) + (nColors * sizeof(TPaletteEntry)));

           lpPalette^.palVersion:=$300;
           lpPalette^.palNumEntries:=nColors;

           byRedMask:=(1 shl pfd.cRedBits)-1;
           byGreenMask:=(1 shl pfd.cGreenBits)-1;
           byBlueMask:=(1 shl pfd.cBlueBits)-1;

           for i:=0 to nColors-1 do
           begin
                lpPalette^.palPalEntry[i].peRed   := (((i shr pfd.cRedShift)   and byRedMask)   * 255) DIV byRedMask;
                lpPalette^.palPalEntry[i].peGreen := (((i shr pfd.cGreenShift) and byGreenMask) * 255) DIV byGreenMask;
                lpPalette^.palPalEntry[i].peBlue  := (((i shr pfd.cBlueShift)  and byBlueMask)  * 255) DIV byBlueMask;
                lpPalette^.palPalEntry[i].peFlags := 0;
           end;

           Palette:=CreatePalette(lpPalette^);
           HeapFree(hHeap,0,lpPalette);

           if Palette<>0 then
           begin
                SelectPalette(DC, Palette, FALSE);
                RealizePalette(DC);
           end;
      end;
      // Restore default flags for the next window or DC.
      DoubleBuffer := TRUE;
      ColorIndex := FALSE;
end;

//This is the function prototype for changing the resolution
//which resides in user32.dll
function ChangeDisplaySettings(lpDevMode: PDeviceModeA; dwFlags: DWORD): Longint; stdcall;
         external user32 name 'ChangeDisplaySettingsA';
// Change the display resolution to rw x rh and rbpp bits.
// Use ChangeRes(0, 0, 0) to restore the normal resolution.
procedure ChangeRes(rw, rh, rbpp: Cardinal);
var
  devMode: TDeviceMode;
  modeExists: LongBool;
  modeSwitch, closeMode, i: Integer;
begin
     closeMode := 0;
     i := 0;
     repeat
           modeExists := EnumDisplaySettings(nil, i, devMode);
           // if not modeExists then: This mode may not be supported. We'll try anyway, though.
           with devMode do
           begin
                if (dmBitsPerPel = rbpp) and (dmPelsWidth = rw) and (dmPelsHeight = rh) then
                begin
                     modeSwitch := ChangeDisplaySettings(@devMode, CDS_FULLSCREEN);
                     if modeSwitch = DISP_CHANGE_SUCCESSFUL then Exit;
                end;
           end;
           if closeMode <> 0 then closeMode := i;
           INC(i);
     until not modeExists;

     EnumDisplaySettings(nil, closeMode, devMode);
     with devMode do
     begin
          dmBitsPerPel := rbpp;
          dmPelsWidth := rw;
          dmPelsHeight := rh;
          dmFields := DM_BITSPERPEL or DM_PELSWIDTH or DM_PELSHEIGHT;
     end;
     modeSwitch := ChangeDisplaySettings(@devMode, CDS_FULLSCREEN);
     if modeSwitch = DISP_CHANGE_SUCCESSFUL then Exit;

     devMode.dmFields := DM_BITSPERPEL;
     modeSwitch := ChangeDisplaySettings(@devMode, CDS_FULLSCREEN);
     if modeSwitch = DISP_CHANGE_SUCCESSFUL then
     begin
          devMode.dmFields := DM_PELSWIDTH or DM_PELSHEIGHT;
          modeSwitch := ChangeDisplaySettings(@devMode, CDS_FULLSCREEN);
          if modeSwitch = DISP_CHANGE_SUCCESSFUL then
          begin
               ChangeDisplaySettings(nil, 0);
               Exit;
          end;
     end;
end;
//set up default params
constructor TOpenGL.Create(aOwner:Tform);
begin
     inherited Create;
     fOwner:=aOwner;
     fXRes:=640;
     fYRes:=480;
end;
//Call CleanUPGL, so that we clean up properly.
destructor TOpenGL.Destroy;
begin
     CleanUpGL;
     inherited Destroy;
end;
//Make our rendering context the current one
procedure TOpenGL.MakeCurrent;
begin
     if (fDC<>0) and (fHRC<>0) then
        wglMakeCurrent(fDC,fHRC);
end;
//InitGL using the given form's DC
procedure TOpenGL.InitGL;
begin
     fDC:=GetDC(fOwner.Handle);
     InitDC(fDC,fPalette);
     fHRC:=wglCreateContext(fDC);
     MakeCurrent;
     fFromHandle:=false;
end;
//InitGL using the DC obtained from the given handle
procedure TOpenGL.InitGLFromHandle(aHandle:THandle);
begin
     fDC:=GetDC(aHandle);
     InitDC(fDC,fPalette);
     fHRC:=wglCreateContext(fDC);
     MakeCurrent;
     fFromHandle:=true;
     fHandle:=aHandle;
end;
//Delete the RC we obtained and release the DC
procedure TOpenGL.CleanUPGL;
begin
     if fFullScreen then
        ChangeRes(0,0,0);
     if fHRC<>0 then
     begin
          wglMakeCurrent(fDC,fHRC);
          wglDeleteContext(fHRC);
     end;
     if fPalette<>0 then DeleteObject(fPalette);
     if not fFromHandle then
        ReleaseDC(fOwner.Handle,fDC)
     else
         ReleaseDC(fHandle,fDC);
end;
//Flip the screen
procedure TOpenGL.Flip;
begin
     SwapBuffers(fDC);
end;
//Set the current clearcolor
procedure TOpenGL.SetClearColor(aColor:TColorf);
begin
     fClearColor:=aColor;
     glClearColor(fClearColor.R,
                  fClearColor.G,
                  fClearColor.B,
                  fClearColor.A);
end;
//Clear the respective buffers
procedure TOpenGL.ClearAll;
begin
     glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT or GL_STENCIL_BUFFER_BIT);
end;
procedure TOpenGL.ClearZBuffer;
begin
     glClear(GL_DEPTH_BUFFER_BIT);
end;
procedure TOpenGL.ClearColorBuffer;
begin
     glClear(GL_COLOR_BUFFER_BIT);
end;
procedure TOpenGL.ClearStencilBuffer;
begin
     glClear(GL_STENCIL_BUFFER_BIT);
end;
//Set fullscreen mode or windowed mode 
procedure TOpenGL.SetFullScreen(v:boolean);
begin
     if v then
     begin
          fFullScreen:=v;
          ChangeRes(fXRes,fYRes,ColorBits);
          fOwner.Left:=0;
          fOwner.Top:=0;
          fOwner.BorderStyle:=bsNone;
          fOwner.WindowState:=wsMaximized;
          fOwner.BringToFront;
     end
     else
         begin
              fFullScreen:=v;
              ChangeREs(0,0,0);
              fOwner.BorderStyle:=bsSingle;
         end;
end;
//Set up default params
constructor TViewVolume.Create;
begin
     fFOV:=30;
     fHither:=1;
     fYon:=100;
end;
//Call gluPerspective to set the projection matrix
//for the suplied params
procedure TViewVolume.SetMatrix;
var
   aspect:single;
begin
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity;
     aspect:=fWidth/fHeight;
     gluPerspective(fFOV,aspect,fHither,fYon);
     glViewport(fX,fY,fWidth,fHeight);
end;

//Get the vendor string
function GetVendor:string;
begin
     result:=string(glGetString(GL_VENDOR));
end;
//Get the renerer string
function GetRenderer:string;
begin
     result:=string(glGetString(GL_RENDERER));
end;
//Get the GL version
function GetGLVersion:string;
begin
     result:=string(glGetString(GL_VERSION));
end;
//Get the GLU version
function GetGLUVersion:string;
begin
     result:=string(glGetString(GLU_VERSION));
end;
//Get the entire list of space separated extensions
function GetExtensions:string;
var
   temp:PChar;
begin
     temp:=PChar(glGetString(GL_EXTENSIONS));
     result:=Copy(temp,0,StrLen(temp));
end;
//Get extensions one by one, parsing the string returned by
//GetExtensions
function GetExtension(num:integer):string;
var
   toknum,startpos,endpos,i:integer;
   s:string;
begin
     s:=GetExtensions;
     toknum:=1;
     startpos:=0;
     endpos:=0;
     for i:=1 to Length(s) do
     begin
          if s[i]=' ' then Inc(toknum);
          if toknum=num then
             startpos:=i+1;
          if toknum=num+1 then
             endpos:=i+1;
     end;
     result:=Copy(s,startpos,endpos-startpos);
end;
//Find the number of extensions by calling GetExtension repeatedly
//until it runs out of extensions
function GetNumExtensions:integer;
var
   count:integer;
   s:string;
begin
     count:=0;
     result:=0;
     while GetExtension(count)<>'' do
     begin
          s:=GetExtension(count);
          if Trim(s)<>'' then
             Inc(count)
          else
          begin
               result:=count;
               exit;
          end;
     end;
end;
end.
