unit Main;

interface

uses
    Windows,
    Messages,
    SysUtils,
    Classes,
    Graphics,
    Controls,
    Forms,
    Dialogs,
    GL,
    GLU,
    GLUT,
    GLClass,
    ExtCtrls;

type
  TMainForm = class(TForm)
    Timer: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    { Private declarations }
  public
    { Public declarations }
    //Declare an instance so we can use it
    GL:TOpenGL;
    //we need to set up a projection matrix, right?
    ViewVolume:TViewVolume;

    //Some mouse movement to add spice to this drab tut
    MousePressed:boolean;
    xStart,xDelta,yStart,yDelta:integer;
    xRot,yRot:single;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

procedure TMainForm.FormCreate(Sender: TObject);
begin
     //Create it. We are going to render to the form.
     GL:=TOpenGL.Create(Self);
     //Check if the user wants to run fullscreen or not.
     GL.FullScreen:=MessageBox(0,'Would you like to run fullscreen?',
                                           'TOpenGL Demo',
                                           MB_YESNO or MB_ICONQUESTION)=ID_YES;
     //That's it. Cool, aint it?
     //Any changes you want to make will never muck up this code.
     GL.InitGL;

     //Create it and set up the width and height of the window
     ViewVolume:=TViewVolume.Create;
     //we could also do this if we wanted to resize the window
     ViewVolume.Width:=ClientWidth;
     ViewVolume.Height:=ClientHeight;

     MousePressed:=false;
     xRot:=0;
     yRot:=0;

     //Let the flipping begin
     Timer.Enabled:=true;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
     //Stop the rendering. We dont want to try and render
     //a destroyed object, right?!!
     Timer.Enabled:=false;

     if assigned(GL) then GL.Destroy;
end;

procedure TMainForm.TimerTimer(Sender: TObject);
begin
     //Clear contents of previous render
     GL.ClearColorBuffer;

     //Set the projection matrix
     ViewVolume.SetMatrix;

     //We will simply translate ourselves and watch a little teapot
     glTranslatef(0,0,-10);
     glRotatef(xRot,1,0,0);
     glRotatef(yRot,0,1,0);
     GLUTWireTeapot(1.0);

     //Flip the screen
     GL.Flip;
end;

procedure TMainForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
     case Key of
          //Now quit, we should...like any good app would :)
          VK_ESCAPE:Close;
     end;
end;

procedure TMainForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
     //Mouse button is being held down
     MousePressed:=true;
     //Note the starting position
     xStart:=X;
     yStart:=Y;
end;

procedure TMainForm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
     //User released the mouse
     MousePressed:=false;
end;

procedure TMainForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
     //Dragging is happening only if the mouse button is down
     if not MousePressed then exit;
     //Calculate how much the mouse has moved
     xDelta:=xStart-X;
     yDelta:=yStart-Y;

     //Adjust rotation and scale so it is'nt too fast
     xRot:=xRot-yDelta/2;
     yrot:=yRot-xDelta/2;

     //Next time, we start from here
     xStart:=X;
     yStart:=Y;
end;

end.
